"""pypdf_table_extraction offers multiple methods to parse pdfs and reconstruct the tables."""

from .hybrid import Hybrid
from .lattice import Lattice
from .network import Network
from .stream import Stream
